@@ -1,6 +1,11 @@
 from __future__ import annotations
-from typing import Dict, Generic, Iterable, Iterator, Union, List, Set, Optional, Tuple, Type, TypeVar, Callable, overload, Literal
-
+from typing import (
+    overload, List, Set, Dict, Generic, Iterable, Iterator, Union, 
+     Optional, Tuple, Type, TypeVar, Callable, Literal, 
+     TYPE_CHECKING
+)
+if TYPE_CHECKING:
+    from random import Random
 
 T = TypeVar("T")
 R = TypeVar("R")
@@ -940,7 +945,7 @@ class Sequence(Generic[T], Iterable[T]):
         """
         return Sequence(self.associate_by(key_selector).values())
     
-    def reduce(self, accumulator: Callable[[T, T], T], initial: Optional[T] = None) -> T:
+    def reduce(self, accumulator: Callable[[R, T], R], initial: Optional[R] = None) -> T:
         """
          Returns the result of applying the specified [accumulator] function to the given Sequence's elements.
 
@@ -958,7 +963,7 @@ class Sequence(Generic[T], Iterable[T]):
             result = accumulator(result, e)
         return result
     
-    def fold(self, initial: R, accumulator: Callable[[R, T], T]) -> R:
+    def fold(self, initial: R, accumulator: Callable[[R, T], R]) -> R:
         """
          Returns the result of applying the specified [accumulator] function to the given Sequence's elements.
 
@@ -1190,6 +1195,18 @@ class Sequence(Generic[T], Iterable[T]):
         return FlatteningSequence(self, transform)
     
 
+    def flatten(self) -> Sequence[T]:
+        """
+         Returns a sequence of all elements from all sequences in this sequence.
+
+         Example 1:
+        >>> lst = [['a', 'b'], ['c'], ['d', 'e']]
+        >>> it(lst).flatten()
+        ['a', 'b', 'c', 'd', 'e']
+        """
+        return FlatteningSequence(self, lambda x: x)
+    
+
     def group_by(self, key_selector: Callable[[T], K]) -> GroupingSequence[K, T]:
         """
          Returns a dictionary with keys being the result of [key_selector] function being invoked on each element of original collection
@@ -1342,6 +1359,82 @@ class Sequence(Generic[T], Iterable[T]):
     def zip(self, other: Iterable[R], transform: Callable[[T, R], V] = lambda a, b:(a, b)) -> Sequence[V]:
         return MergingSequence(self, other, transform)
     
+    def zip_with_next(self, transform: Callable[[T, T], V]) -> Sequence[V]:
+        """
+         Returns a sequence containing the results of applying the given [transform] function
+        to an each pair of two adjacent elements in this sequence.
+        
+         Example 1:
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).zip_with_next(lambda x, y: x + '__' + y)
+        ['a__b', 'b__c']
+        """
+        return MergingWithNextSequence(self, transform)
+    
+    def with_index(self):
+        """
+         Returns a sequence containing the elements of this sequence and their indexes.
+        
+         Example 1:
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).with_index()
+        [IndexedValue(0, a), IndexedValue(1, b), IndexedValue(2, c)]
+        """
+        return IndexingSequence(self)
+    
+
+    @overload
+    def shuffled(self) -> Sequence[T]:
+        """
+         Returns a sequence that yields elements of this sequence randomly shuffled
+        using the specified [random] instance as the source of randomness.
+
+         Example 1:
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).shuffled('123')
+        ['b', 'a', 'c']
+
+         Example 2:
+        >>> from random import Random
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).shuffled(Random('123'))
+        ['b', 'a', 'c']
+
+         Example 3:
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).shuffled(123)
+        ['c', 'b', 'a']
+        """
+        ...
+    @overload
+    def shuffled(self, random: Random) -> Sequence[T]:
+        ...
+    @overload
+    def shuffled(self, seed: Union[int, str]) -> Sequence[T]:
+        ...
+    def shuffled(self, random: Optional[Union[Random, str, int]]=None) -> Sequence[T]:
+        """
+         Returns a sequence that yields elements of this sequence randomly shuffled
+        using the specified [random] instance as the source of randomness.
+
+         Example 1:
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).shuffled('123')
+        ['b', 'a', 'c']
+
+         Example 2:
+        >>> from random import Random
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).shuffled(Random('123'))
+        ['b', 'a', 'c']
+
+         Example 3:
+        >>> lst = ['a', 'b', 'c']
+        >>> it(lst).shuffled(123)
+        ['c', 'b', 'a']
+        """
+        return ShufflingSequence(self, random)
+    
     def partition(self, predicate: Callable[[T], bool]) -> Tuple[Sequence[T], Sequence[T]]:
         """
          Partitions the elements of the given Sequence into two groups,
@@ -1710,6 +1803,25 @@ class MergingSequence(Sequence[V]):
                 break
 
 
+class MergingWithNextSequence(Sequence[V]):
+    def __init__(self, 
+    iterable: Iterable[T], 
+    transformer: Callable[[T, R], V] = lambda a,b:(a,b)) -> None:
+        super().__init__(iterable)
+        self._transformer = transformer
+    
+    def __do_iter__(self) -> Iterator[V]:
+        it = iter(self._iter)
+        try:
+            c = next(it)
+            while True:
+                n = next(it)
+                yield self._transformer(c, n)
+                c = n
+        except StopIteration:
+            pass
+
+
 class GroupingSequence(Sequence[Tuple[K, List[T]]]):
     def __init__(self, iterable: Iterable[T], key_func: Callable[[T], K]) -> None:
         super().__init__(iterable)
@@ -1776,6 +1888,21 @@ class ConcatSequence(Sequence[T]):
             yield from i
 
 
+class ShufflingSequence(Sequence[T]):
+    def __init__(self, iterable: Iterable[T], random: Optional[Union[Random, str, int]]=None) -> None:
+        super().__init__(iterable)
+        if random is None or isinstance(random, (str, int)):
+            from random import Random
+            self._random = Random(random)
+        else:
+            self._random = random
+    
+    def __do_iter__(self) -> Iterator[T]:
+        lst = self._iter._cache.copy() if isinstance(self._iter, Sequence) and self._iter._cache is not None else list(self._iter)
+        self._random.shuffle(lst)
+        yield from lst
+
+
 class ProgressSequence(Sequence[T]):
     def __init__(self, iterable: Sequence[T], progress_func: Callable[[Iterable[T]], Iterable[T]]) -> None:
         super().__init__(iterable)
